package reportclient

import (
	"fmt"
	"net/http"
	"net/url"
	"encoding/json"
	"io/ioutil"
	"os"
	"strings"
	"bytes"
	"io"
	"mime/multipart"
	_ "crypto/sha512"
)

var (
	ReportsUri = "https://reportsapi.zoho.com"
	AccountsUri = "https://accounts.zoho.com"
	RequestUri string 
	Action string
	OutputFormat = "JSON"
	ErrorFormat = "JSON"
	AuthToken string
	ApiVersion = "1.0"
	Importfilepath string
	Exportfilename string
	IsOAuth = false
)

/*
SetAuthToken - Set an authtoken for the specified user.

authtoken - The authtoken for the user. 
*/
func SetAuthToken(authtoken string) {
	AuthToken = authtoken
}

/*
SetOAuthToken - Set an OAuthtoken for the specified user.

clientid - The clientid for the user. 

clientsecret - The clientsecret for the user. 

refreshtoken - The refreshtoken for the user. 
*/
func SetOAuthToken(clientid, clientsecret, refreshtoken string) {
	params := map[string]string{}
	params["client_id"] = clientid
	params["client_secret"] = clientsecret
	params["refresh_token"] = refreshtoken
	params["grant_type"] = "refresh_token"
	
	AuthToken = GetOAuthToken(params)
	IsOAuth = true
}

/*
AddRow - Adds row to the specified table identified by the URI.

tableuri - The URI of the table. 

columnvalues - Contains the values for the row. The column name(s) are the key.

addedrows - Successfully added rows with value.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func AddRow(tableuri string, columnvalues map[string]string) (addedrows interface{}, SeErr *ServerException) {
	RequestUri = tableuri
	Action = "ADDROW"
	addedrows , SeErr = SendRequest(columnvalues, true)
	return
}

/*
DeleteData - Delete the data in the specified table identified by the URI.

tableuri - The URI of the table. 

params - Contains any additional control parameters. Can be null.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func DeleteData(tableuri string, params map[string]string) (SeErr *ServerException) {
	RequestUri = tableuri
	Action = "DELETE"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
UpdateData - Update the data in the specified table identified by the URI.

tableuri - The URI of the table. 

params - Contains any additional control parameters. Can be null.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func UpdateData(tableuri string, params map[string]string) (SeErr *ServerException) {
	RequestUri = tableuri
	Action = "UPDATE"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
ImportData - Import the data contained in a given file into the table identified by the URI.

tableuri - The URI of the table. 

file - The file containing the data to be imported into the table.

importtype - The type of import.

autoidentity - Used to specify whether to auto identify the CSV format.

onerror - This parameter controls the action to be taken incase there is an error during import.

params - Contains any additional control parameters. Can be null.

ImportResult - Structure pointer for the import result.

ServerException - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func ImportData(tableuri, file, importtype, autoidentity, onerror string, params map[string]string) (ImportResult, *ServerException) {
	Importfilepath = file
	params["ZOHO_IMPORT_TYPE"] = importtype
	params["ZOHO_AUTO_IDENTIFY"] = autoidentity
	params["ZOHO_ON_IMPORT_ERROR"] = onerror
	RequestUri = tableuri
	Action = "IMPORT"
	response  , SeErr := SendImportRequest(params)
	imp := &ImportResult{}
	if(SeErr == nil){
		imp.ImportErrors = response.(map[string]interface{})["importErrors"].(string)
		imp.ColumnDetails = response.(map[string]interface{})["columnDetails"]
		importSummary := response.(map[string]interface{})["importSummary"]
		imp.ImportType = importSummary.(map[string]interface{})["importType"].(string)
		imp.Warnings = importSummary.(map[string]interface{})["warnings"].(float64)
		imp.SelectedColumnCount = importSummary.(map[string]interface{})["selectedColumnCount"].(float64)
		imp.SuccessRowCount = importSummary.(map[string]interface{})["successRowCount"].(float64)
		imp.ImportOperation = importSummary.(map[string]interface{})["importOperation"].(string)
		imp.TotalColumnCount = importSummary.(map[string]interface{})["totalColumnCount"].(float64)
		imp.TotalRowCount = importSummary.(map[string]interface{})["totalRowCount"].(float64)
	}
	return *imp , SeErr
}

/*
ExportData - Exports the data/report of table (or report) identified by the URI.

tableuri - The URI of the table. 

file - The name of the file.

outputformat - The format in which the data is to be exported.

params - Contains any additional control parameters. Can be null.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func ExportData(tableuri, file, outputformat string, params map[string]string) (SeErr *ServerException) {
	Exportfilename = file
	RequestUri = tableuri
	OutputFormat = outputformat
	Action = "EXPORT"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
CopyDatabase - Copy a specified database identified by the URI.

dburi - The URI of the database. 

newdbname - Contains new database name.

dbkey - Contains database key of the source database.

params - Contains any additional control parameters. Can be null.

newdbid - The new database id.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func CopyDatabase(dburi, newdbname, dbkey string, params map[string]string ) (newdbid string, SeErr *ServerException) {
	params["ZOHO_DATABASE_NAME"] = newdbname
	params["ZOHO_COPY_DB_KEY"] = dbkey
	RequestUri = dburi
	Action = "COPYDATABASE"
	response , exception := SendRequest(params, true)
	if(exception != nil){
		SeErr = exception
	}else{
		newdbid = response.(map[string]interface{})["dbid"].(string)
	}
	return
}

/*
DeleteDatabase - Delete a specified database from a Zoho Reports Account.

useruri - The URI of the user. 

dbname - The name of the database to be deleted from the Zoho Reports Account.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func DeleteDatabase(useruri, dbname string) (SeErr *ServerException) {
	params := map[string]string{}
	params["ZOHO_DATABASE_NAME"] = dbname
	RequestUri = useruri
	Action = "DELETEDATABASE"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
EnableDomainDB - Enable database for custom domain..

useruri - The URI of the user.

dbname - The database names which you want to show in your custom domain.

domainname - Custom domain name.

params - Can contain any additional params.

result - Returns Generated result as interface.

ServerException - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func EnableDomainDB(useruri, dbname, domainname string, params map[string]string) (result interface{}, SeErr *ServerException){
	RequestUri = useruri
	params["DBNAME"] = dbname
	params["DOMAINNAME"] = domainname
	Action = "ENABLEDOMAINDB"
	result , SeErr = SendRequest(params, true)
	return
}

/*
DisableDomainDB - Disable database in custom domain..

useruri - The URI of the user.

dbname - The database names which you want to disable from your custom domain.

domainname - Custom domain name.

params - Can contain any additional params.

result - Returns Generated result as interface.

ServerException - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func DisableDomainDB(useruri, dbname, domainname string, params map[string]string) (result interface{}, SeErr *ServerException){
	RequestUri = useruri
	params["DBNAME"] = dbname
	params["DOMAINNAME"] = domainname
	Action = "DISABLEDOMAINDB"
	result , SeErr = SendRequest(params, true)
	return
}

/*
CreateTable - Create a table in the specified database.

dburi - The URI of the database. 

tableconfig - Table structure in JSON format (includes table name, description, folder name, column and lookup details, is system table).

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func CreateTable(dburi, tableconfig string) (SeErr *ServerException) {
	params := map[string]string{}
	params["ZOHO_TABLE_DESIGN"] = tableconfig
	RequestUri = dburi
	Action = "CREATETABLE"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
AutoGenReports - Generate reports for table or column.

tableuri - The URI of the table.

source - Should be column or table.

params - Can contain any additional params.

result - Returns Generated result as interface.

ServerException - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func AutoGenReports(tableuri string, source string, params map[string]string) (result interface{}, SeErr *ServerException) {
	params["ZOHO_SOURCE"] = source
	RequestUri = tableuri
	Action = "AUTOGENREPORTS"
	result , SeErr = SendRequest(params, true)
	return
}

/*
CreateAnalysisView - Create a report in the specified database.

dburi - The URI of the database. 

reportconfig - Report structure in JSON format.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func CreateAnalysisView(dburi, reportconfig string) (SeErr *ServerException) {
	params := map[string]string{}
	params["ZOHO_VIEW_DATA"] = reportconfig
	RequestUri = dburi
	Action = "CREATEANALYSISVIEW"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
CreateSimilarViews - Create reports similar as another table reports.

tableuri - The URI of the table.

refview - The reference table name.

foldername - Folder name where the reports to be saved.

customformula - If true, it will create reports with custom formula else it will ignore that formula.

aggformula - If true, it will create reports with aggregate formula else it will ignore that formula.

params - Can contain any additional params.

result - Returns Generated result as interface.

ServerException - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func CreateSimilarViews(tableuri, refview, foldername string, customformula, aggformula bool, params map[string]string) (result interface{}, SeErr *ServerException){
	RequestUri = tableuri
	params["ZOHO_REFVIEW"] = refview
	params["ZOHO_FOLDERNAME"] = foldername
	params["ISCOPYCUSTOMFORMULA"] = "false"
	if(customformula == true){
		params["ISCOPYCUSTOMFORMULA"] = "true"
	}
	params["ISCOPYAGGFORMULA"] = "false"
	if(aggformula == true){
		params["ISCOPYAGGFORMULA"] = "true"
	}
	Action = "CREATESIMILARVIEWS"
	result , SeErr = SendRequest(params, true)
	return
}

/*
RenameView - Rename the specified view with the new name and description.

dburi - The URI of the database. 

viewname - Current name of the view.

newviewname - New name for the view.

newdesc - New description for the view.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func RenameView(dburi, viewname, newviewname, newdesc string) (SeErr *ServerException) {
	params := map[string]string{}
	params["ZOHO_VIEWNAME"] = viewname
	params["ZOHO_NEW_VIEWNAME"] = newviewname
	params["ZOHO_NEW_VIEWDESC"] = newdesc
	RequestUri = dburi
	Action = "RENAMEVIEW"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
DeleteView - Delete the specified view from the database.

dburi - The URI of the database. 

viewname - View name to be deleted.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func DeleteView(dburi, viewname string) (SeErr *ServerException) {
	params := map[string]string{}
	params["ZOHO_VIEW"] = viewname
	RequestUri = dburi
	Action = "DELETEVIEW"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
DuplicateViews - Copy views within the database(Save as).

dburi - The URI of the database.

viewtocopy - The reference view name.

newviewname - New view name.

params - Can contain any additional params.

result - Duplicate views result.

ServerException - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func DuplicateViews(dburi, viewtocopy, newviewname string, params map[string]string) (result string, SeErr *ServerException){
	RequestUri = dburi
	params["ZOHO_VIEWTOCOPY"] = viewtocopy
	params["ZOHO_NEW_VIEWNAME"] = newviewname
	Action = "DUPLICATEVIEWS"
	response , exception := SendRequest(params, true)
	if(exception != nil){
		SeErr = exception
	}else{
		result = response.(map[string]interface{})["message"].(string)
	}
	return
}

/*
CopyReports - The Copy Reports API is used to copy one or more reports from one database to another within the same account or even across user accounts.

dburi - The URI of the database. 

viewname - This parameter holds the list of view names.

dbname - The database name where the report's had to be copied.

dbkey - The secret key used for allowing the user to copy the reports.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func CopyReports(dburi, viewname, dbname, dbkey string) (SeErr *ServerException) {
	params := map[string]string{}
	params["ZOHO_VIEWTOCOPY"] = viewname
	params["ZOHO_DATABASE_NAME"] = dbname
	params["ZOHO_COPY_DB_KEY"] = dbkey
	RequestUri = dburi
	Action = "COPYREPORTS"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
CopyFormula - The Copy Formula API is used to copy one or more formula columns from one table to another within the same database or across databases and even across one user account to another.

tableuri - The URI of the table. 

formulaname - This parameter holds the list of formula names.

dbname - The database name where the formula's had to be copied.

dbkey - The secret key used for allowing the user to copy the formula.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func CopyFormula(tableuri, formulaname, dbname, dbkey string) (SeErr *ServerException) {
	params := map[string]string{}
	params["ZOHO_FORMULATOCOPY"] = formulaname
	params["ZOHO_DATABASE_NAME"] = dbname
	params["ZOHO_COPY_DB_KEY"] = dbkey
	RequestUri = tableuri
	Action = "COPYFORMULA"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
AddColumn - Adds a column to the specified table identified by the URI.

tableuri - The URI of the table. 

columnname - Contains the name of the column to be added.

datatype - Contains the datatype of the column to be added.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func AddColumn(tableuri, columnname, datatype string) (SeErr *ServerException) {
	params := map[string]string{}
	params["ZOHO_COLUMNNAME"] = columnname
	params["ZOHO_DATATYPE"] = datatype
	RequestUri = tableuri
	Action = "ADDCOLUMN"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
DeleteColumn - Delete the column in the specified table identified by the URI.

tableuri - The URI of the table. 

columnname - Contains the name of the column to be deleted.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func DeleteColumn(tableuri, columnname string) (SeErr *ServerException) {
	params := map[string]string{}
	params["ZOHO_COLUMNNAME"] = columnname
	RequestUri = tableuri
	Action = "DELETECOLUMN"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
RenameColumn - Rename the column in the specified table identified by the URI.

tableuri - The URI of the table. 

olcolumnname - Contains the name of the column to be modified.

newcolumnname - Contains the new column name.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func RenameColumn(tableuri, olcolumnname, newcolumnname string) (SeErr *ServerException) {
	params := map[string]string{}
	params["OLDCOLUMNNAME"] = olcolumnname
	params["NEWCOLUMNNAME"] = newcolumnname
	RequestUri = tableuri
	Action = "RENAMECOLUMN"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
HideColumn - Hide the specified columns in the particular table.

tableuri - The URI of the table.

columnnames - Contains array of column names.

params - Can contain any additional params.

result - Returns Generated result as interface.

ServerException - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func HideColumn(tableuri string, columnnames []string, params map[string]string) (result interface{}, SeErr *ServerException){
	RequestUri = tableuri
	for _,columnname := range columnnames{
		RequestUri += "&ZOHO_COLUMNNAME=" + columnname
	}
	RequestUri = RequestUri + "&"
	Action = "HIDECOLUMN"
	result , SeErr = SendRequest(params, true)
	return
}

/*
ShowColumn - Show the hidden columns in the particular table.

tableuri - The URI of the table.

columnnames - Contains array of column names.

params - Can contain any additional params.

result - Returns Generated result as interface.

ServerException - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func ShowColumn(tableuri string, columnnames []string, params map[string]string) (result interface{}, SeErr *ServerException){
	RequestUri = tableuri
	for _,columnname := range columnnames{
		RequestUri += "&ZOHO_COLUMNNAME=" + columnname
	}
	RequestUri = RequestUri + "&"
	Action = "SHOWCOLUMN"
	result , SeErr = SendRequest(params, true)
	return
}

/*
AddLookup - Add the lookup for the given column.

tableuri - The URI of the table. 

columnname - Name of the column (Child column).

refertable - Name of the referred table (parent table).

refercolumn - Name of the referred column (parent column).

onerror - This parameter controls the action to be taken incase there is an error during lookup.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func AddLookup(tableuri, columnname, refertable, refercolumn, onerror string) (SeErr *ServerException) {
	params := map[string]string{}
	params["ZOHO_COLUMNNAME"] = columnname
	params["ZOHO_REFERREDTABLE"] = refertable
	params["ZOHO_REFERREDCOLUMN"] = refercolumn
	params["ZOHO_IFERRORONCONVERSION"] = onerror
	RequestUri = tableuri
	Action = "ADDLOOKUP"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
RemoveLookup - Remove the lookup for the given column.

tableuri - The URI of the table. 

columnname - Name of the column.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func RemoveLookup(tableuri, columnname string) (SeErr *ServerException) {
	params := map[string]string{}
	params["ZOHO_COLUMNNAME"] = columnname
	RequestUri = tableuri
	Action = "REMOVELOOKUP"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
GetDatabaseMetadata - This method is used to get the meta information about the reports.

uri - The URI of the user or database. 

metadata - It specifies the information to be fetched.

metainfo - Contains the metadata.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func GetDatabaseMetadata(uri, metadata string) (metainfo interface{}, SeErr *ServerException) {
	params := map[string]string{}
	params["ZOHO_METADATA"] = metadata
	RequestUri = uri
	Action = "DATABASEMETADATA"
	metainfo , SeErr = SendRequest(params, true)
	return
}

/*
GetDatabaseName - Get database name for a specified database identified by the URI.

useruri - The URI of the user. 

dbid - The ID of the database.

databasename - Contains the Database name.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func GetDatabaseName(useruri, dbid string) (databasename string, SeErr *ServerException) {
	params := map[string]string{}
	params["DBID"] = dbid
	RequestUri = useruri
	Action = "GETDATABASENAME"
	response , exception := SendRequest(params, true)
	if(exception != nil){
		SeErr = exception
	}else{
		databasename = response.(map[string]interface{})["dbname"].(string)
	}
	return
}

/*
IsDbExist - Used to check wheather the database is exist or not.

useruri - The URI of the user. 

dbname - The database name.

params - Can contain any additional params.

result - Returns isdbexist result.

ServerException - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func IsDbExist(useruri string, dbname string, params map[string]string) (result string, SeErr *ServerException) {
	params["ZOHO_DB_NAME"] = dbname
	RequestUri = useruri
	Action = "ISDBEXIST"
	response , exception := SendRequest(params, true)
	if(exception != nil){
		SeErr = exception
	}else{
		result = response.(map[string]interface{})["isdbexist"].(string)
	}
	return
}

/*
GetCopyDbKey - Get database key for the particular database.

dburi - The URI of the database. 

dbkey - Returns the database key.

ServerException - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func GetCopyDbKey(dburi string) (dbkey string, SeErr *ServerException){
	params := map[string]string{}
	RequestUri = dburi
	Action = "GETCOPYDBKEY"
	response , exception := SendRequest(params, true)
	if(exception != nil){
		SeErr = exception
	}else{
		dbkey = response.(map[string]interface{})["copydbkey"].(string)
	}
	return
}

/*
GetViewname - This function returns the name of a view in Zoho Reports.

useruri - The URI of the user. 

objid - The view id (object id).

viewname - Contains the View name.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func GetViewname(useruri, objid string) (viewname string, SeErr *ServerException) {
	params := map[string]string{}
	params["OBJID"] = objid
	RequestUri = useruri
	Action = "GETVIEWNAME"
	response , exception := SendRequest(params, true)
	if(exception != nil){
		SeErr = exception
	}else{
		viewname = response.(map[string]interface{})["viewname"].(string)
	}
	return
}

/*
GetInfo - This method returns the Database ID (DBID) and View ID (OBJID) of the corresponding Database.

tableuri - The URI of the table. 

info - Contains the View-Id (object id) and Database-Id.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func GetInfo(tableuri string) (info interface{}, SeErr *ServerException) {
	params := map[string]string{}
	RequestUri = tableuri
	Action = "GETINFO"
	info , SeErr = SendRequest(params, true)
	return
}

/*
ShareView - This method is used to share the views (tables/reports/dashboards) created in Zoho Reports with users.

dburi - The URI of the database. 

emailIds - It contains the owners email-id.

views - It contains the view names.

params - Contains any additional control parameters. Can be null.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func ShareView(dburi, emailIds , views string, params map[string]string) (SeErr *ServerException) {
	params["ZOHO_EMAILS"] = emailIds
	params["ZOHO_VIEWS"] = views
	RequestUri = dburi
	Action = "SHARE"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
RemoveShare - This method is used to remove the shared views (tables/reports/dashboards) in Zoho Reports from the users.

dburi - The URI of the database. 

emailIds - It contains the owners email-id.

params - Contains any additional control parameters. Can be null.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func RemoveShare(dburi, emailIds string, params map[string]string) (SeErr *ServerException) {
	params["ZOHO_EMAILS"] = emailIds
	RequestUri = dburi
	Action = "REMOVESHARE"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
AddDbOwner - This method is used to add new users as owner to the reports database.

dburi - The URI of the database. 

emailIds - It contains the users email-id.

params - Contains any additional control parameters. Can be null.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func AddDbOwner(dburi, emailIds string, params map[string]string) (SeErr *ServerException) {
	params["ZOHO_EMAILS"] = emailIds
	RequestUri = dburi
	Action = "ADDDBOWNER"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
RemoveDbOwner - This method is used to remove the existing owners from the reports database.

dburi - The URI of the database. 

emailIds - It contains the users email-id.

params - Contains any additional control parameters. Can be null.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func RemoveDbOwner(dburi, emailIds string, params map[string]string) (SeErr *ServerException) {
	params["ZOHO_EMAILS"] = emailIds
	RequestUri = dburi
	Action = "REMOVEDBOWNER"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
GetShareInfo - Get the sharing information of the database.

dburi - The URI of the database. 

ShareInfo - Returns shareinfo structure pointer.

ServerException - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func GetShareInfo(dburi string) (ShareInfo , *ServerException) {
	params := map[string]string{}
	RequestUri = dburi
	Action = "GETSHAREINFO"
	response , SeErr := SendRequest(params, true)
	info := &ShareInfo{}
	if(SeErr == nil){
		info.AdminMembers = response.(map[string]interface{})["dbownershareinfo"].(map[string]interface{})["dbowners"]

		userinfo := response.(map[string]interface{})["usershareinfo"]
		if(userinfo != nil){	
			var shareduser []string
			for _ , ele := range userinfo.([]interface{}) {
				for _ , ele2 := range ele.(map[string]interface{}) {
					userid := ele2.(map[string]interface{})["email"].(string)
					shareduser = append(shareduser,userid)
				}
			}
			info.SharedUsers = shareduser
			info.UserInfo = Getkeyinfo(userinfo, "email")
		}

		groupinfo := response.(map[string]interface{})["groupshareinfo"]
		if(groupinfo != nil){
			groupmembers := make(map[int]interface{})
			i := 0
			for _ , ele := range groupinfo.([]interface{}) {
				for _ , ele2 := range ele.(map[string]interface{}) {
					groupname := ele2.(map[string]interface{})["groupName"].(string)
					groupdesc := ele2.(map[string]interface{})["desc"].(string)
					members := ele2.(map[string]interface{})["groupmembers"]
					grpmembers := make(map[string]interface{})
					grpmembers["name"] = groupname
					grpmembers["desc"] = groupdesc
					grpmembers["members"] = members
					groupmembers[i] = grpmembers
					i++
				}
			}
			info.GroupMembers = groupmembers
			info.GroupInfo = Getkeyinfo(groupinfo, "group")
		}

		publicinfo := response.(map[string]interface{})["publicshareinfo"]
		if(publicinfo != nil){
			info.PublicInfo = Getinfo(publicinfo)
		}

		privateinfo := response.(map[string]interface{})["privatelinkshareinfo"]
		if(privateinfo != nil){
			info.PrivateInfo = Getinfo(privateinfo)
		}
	}
	return *info , SeErr
}

/*
Internally used.
*/
func Getinfo(resp interface{}) (interface{}) {
	shareinfo := make(map[string]interface{})
	userid := resp.(map[string]interface{})["email"].(string)
	perminfo := make(map[string]interface{})
	for _ , ele := range resp.(map[string]interface{})["permissions"].([]interface{}) {
		permission := make(map[string]interface{})
		viewname := ele.(map[string]interface{})["perminfo"].(map[string]interface{})["viewname"].(string)
		sharedby := ele.(map[string]interface{})["perminfo"].(map[string]interface{})["sharedby"].(string)
		permissions := ele.(map[string]interface{})["perminfo"].(map[string]interface{})["permission"]
		permission["sharedby"] = sharedby
		permission["permissions"] = permissions
		perminfo[viewname] = permission
	}
	shareinfo[userid] = perminfo
	return shareinfo
}

/*
Internally used.
*/
func Getkeyinfo(resp interface{}, key string) (interface{}){
	shareinfo := make(map[string]interface{})
	var userid string
	var info interface{}
	for _ , ele1 := range resp.([]interface{}) {
		for _ , ele2 := range ele1.(map[string]interface{}) {
			if(key == "email"){
				userid = ele2.(map[string]interface{})["email"].(string)
				info = ele2.(map[string]interface{})["permissions"]
			}else{
				userid = ele2.(map[string]interface{})["groupName"].(string)
				info = ele2.(map[string]interface{})["permissions"]
			}
			perminfo := make(map[string]interface{})
			for _ , ele3 := range info.([]interface{}) {
				permission := make(map[string]interface{})
				viewname := ele3.(map[string]interface{})["perminfo"].(map[string]interface{})["viewname"].(string)
				sharedby := ele3.(map[string]interface{})["perminfo"].(map[string]interface{})["sharedby"].(string)
				permissions := ele3.(map[string]interface{})["perminfo"].(map[string]interface{})["permission"]
				permission["sharedby"] = sharedby
				permission["permissions"] = permissions
				perminfo[viewname] = permission
			}
			shareinfo[userid] = perminfo
		}
	}
	return shareinfo
}

/*
GetViewUrl - This method returns the URL to access the mentioned view.

tableuri - The URI of the table. 

viewurl - The view URI.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func GetViewUrl(tableuri string) (viewurl string, SeErr *ServerException) {
	params := map[string]string{}
	RequestUri = tableuri
	Action = "GETVIEWURL"
	response , exception := SendRequest(params, true)
	if(exception != nil){
		SeErr = exception
	}else{
		viewurl = response.(map[string]interface{})["viewurl"].(string)
	}
	return
}

/*
GetEmbedUrl - This method is used to get the embed URL of the particular table / view. This API is available only for the White Label Administrator. 

tableuri - The URI of the table. 

params - Contains any additional control parameters. Can be null.

embedurl - The embed URI.

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func GetEmbedUrl(tableuri string, params map[string]string) (embedurl string, SeErr *ServerException) {
	RequestUri = tableuri
	Action = "GETEMBEDURL"
	response , exception := SendRequest(params, true)
	if(exception != nil){
		SeErr = exception
	}else{
		embedurl = response.(map[string]interface{})["embedUrl"].(string)
	}
	return
}

/*
GetUsers - Return all the users list for that account.

useruri - The URI of the user.

params - Can contain any additional params.

userslist - Returns userslist as interface.

ServerException - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func GetUsers(useruri string, params map[string]string) (userslist interface{}, SeErr *ServerException) {
	RequestUri = useruri
	Action = "GETUSERS"
	userslist , SeErr = SendRequest(params, true)
	return
}

/*
AddUser - Add the specified user(s) into your Zoho Reports Account. 

useruri - The URI of the user. 

emailIds - The email addresses of the users to be added to your Zoho Reports Account separated by comma. 

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func AddUser(useruri, emailIds string) (SeErr *ServerException) {
	params := map[string]string{}
	params["ZOHO_EMAILS"] = emailIds
	RequestUri = useruri
	Action = "ADDUSER"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
RemoveUser - Remove the specified user(s) from your Zoho Reports Account. 

useruri - The URI of the user. 

emailIds - The email addresses of the users to be removed from your Zoho Reports Account separated by comma. 

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func RemoveUser(useruri, emailIds string) (SeErr *ServerException) {
	params := map[string]string{}
	params["ZOHO_EMAILS"] = emailIds
	RequestUri = useruri
	Action = "REMOVEUSER"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
ActivateUser - Activate the specified user(s) in your Zoho Reports Account. 

useruri - The URI of the user. 

emailIds - The email addresses of the users to be activated in your Zoho Reports Account separated by comma. 

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func ActivateUser(useruri, emailIds string) (SeErr *ServerException) {
	params := map[string]string{}
	params["ZOHO_EMAILS"] = emailIds
	RequestUri = useruri
	Action = "ACTIVATEUSER"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
DeActivateUser - Deactivate the specified user(s) in your Zoho Reports Account. 

useruri - The URI of the user. 

emailIds - The email addresses of the users to be deactivated in your Zoho Reports Account separated by comma. 

SeErr - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func DeActivateUser(useruri, emailIds string) (SeErr *ServerException) {
	params := map[string]string{}
	params["ZOHO_EMAILS"] = emailIds
	RequestUri = useruri
	Action = "DEACTIVATEUSER"
	_ , SeErr = SendRequest(params, false)
	return
}

/*
GetPlanInfo - Get the plan information.

useruri - The URI of the user. 

PlanInfo - Returns planinfo structure pointer.

ServerException - Throws a server exception. If the server has recieved the request but did not process the request due to some error.
*/
func GetPlanInfo(useruri string) (PlanInfo, *ServerException){
	params := map[string]string{}
	RequestUri = useruri
	Action = "GETUSERPLANDETAILS"
	response , SeErr := SendRequest(params, true)
	info := &PlanInfo{}
	if(SeErr == nil){
		info.Plan = response.(map[string]interface{})["plan"].(string)
		info.Addons = response.(map[string]interface{})["addon"].(string)
		info.BillingDate = response.(map[string]interface{})["billingDate"].(string)
		info.RowsAllowed = response.(map[string]interface{})["rowsAllowed"].(string)
		info.RowsUsed = response.(map[string]interface{})["rowsUsed"].(string)
		info.TrialAvailed = response.(map[string]interface{})["TrialAvailed"].(string)
		if(info.TrialAvailed != "false"){
			info.TrialPlan = response.(map[string]interface{})["TrialPlan"].(string)
			info.TrialStatus = response.(map[string]interface{})["TrialStatus"].(string)
			info.TrialEndDate = response.(map[string]interface{})["TrialEndDate"].(string)
		}
	}
	return *info , SeErr
}

/*
GetUserUri - Returns the URI for the specified user login email id.

emailid - Emailid of the user.
*/
func GetUserUri(emailid string) string{
	return ReportsUri + "/api/" + emailid + "?"
}

/*
GetDbUri - Returns the URI for the specified database in the ZohoReports account.

emailid - Emailid of the user.

dbname - Name of the database.
*/
func GetDbUri(emailid , dbname string) string{
	return ReportsUri + "/api/" + emailid + "/" + UrlSplCharReplace(dbname) + "?"
}

/*
GetUri - Returns the URI for the specified table in the ZohoReports account.

emailid - Emailid of the user.

dbname - Name of the database.

tbname - Name of the table.
*/
func GetUri(emailid , dbname , tbname string) string{
	return ReportsUri + "/api/" + emailid + "/" + UrlSplCharReplace(dbname) + "/" + UrlSplCharReplace(tbname) + "?"
}

/*
Internally used. For handling special char's in the database name or table name.
*/
func UrlSplCharReplace(value string) string{
	value = strings.Replace(value, "/", "(/)", -1);
	value = strings.Replace(value, "\\", "(//)", -1);
	return value
}

/*
SendImportRequest - Internally used for Import data.

Send request to the server and get response.
*/
func SendImportRequest(config map[string]string) (msg interface{}, SeErr *ServerException) {
	params := url.Values{
		"ZOHO_ACTION": {Action},
		"ZOHO_OUTPUT_FORMAT": {OutputFormat},
		"ZOHO_ERROR_FORMAT": {ErrorFormat},
		"ZOHO_API_VERSION":  {ApiVersion},
	}

	if IsOAuth == false {
		params.Add("authtoken",AuthToken)		
	}

	for key, value := range config {
		params.Add(key,value)
	}

	_, key := params["ZOHO_CREATE_TABLE"]

	if(key == false){
		value := "false"
		params.Add("ZOHO_CREATE_TABLE",value)
	}

	param := params.Encode()
    	FinalURL := RequestUri+param

	bodyBuf := bytes.NewBufferString("")
	bodyWriter := multipart.NewWriter(bodyBuf)
	_, err := bodyWriter.CreateFormFile("ZOHO_FILE", Importfilepath)
	if err != nil {
		panic(err)
	}
	file, err := os.Open(Importfilepath)
	if err != nil {
		panic(err)
	} 
	boundary := bodyWriter.Boundary()
	closeBuf := bytes.NewBufferString(fmt.Sprintf("\r\n--%s--\r\n", boundary))
	requestReader := io.MultiReader(bodyBuf, file, closeBuf)
	filereq, err := file.Stat()
	if err != nil {
		panic(err)
	}

	request, err := http.NewRequest("POST", FinalURL, requestReader)
	if err != nil {
		panic(err)
	}

	if IsOAuth == true {
		request.Header.Set("Authorization" , "Zoho-oauthtoken "+AuthToken)
	}
	
	request.Header.Add("Content-Type", "multipart/form-data; boundary="+boundary)
	request.ContentLength = filereq.Size() + int64(bodyBuf.Len()) + int64(closeBuf.Len())

	client := &http.Client{}
	response, resperror := client.Do(request)
    	if resperror != nil {
        		panic(resperror)
    	}

    	defer response.Body.Close()
    	respbody, _ := ioutil.ReadAll(response.Body)

	var jsonresult map[string]interface{}

    	if(response.StatusCode == 200){
    		json.Unmarshal(respbody, &jsonresult)
    		msg = jsonresult["response"].(map[string]interface{})["result"]
    		return
    	}else {
    		json.Unmarshal(respbody, &jsonresult)
    		excep := &ServerException{}
    		SeErr = excep
    		SeErr.ErrorMessage = jsonresult["response"].(map[string]interface{})["error"].(map[string]interface{})["message"].(string)
    		SeErr.ErrorCode = jsonresult["response"].(map[string]interface{})["error"].(map[string]interface{})["code"].(float64)
    		SeErr.HttpStatusCode = response.StatusCode
    		SeErr.Action = Action
    		return
    	}
}

/*
GetOAuthToken - Internal method for getting OAuth token.
*/
func GetOAuthToken(params map[string]string) (token string){
	
	urlparams := url.Values{}

	for key, value := range params {
		urlparams.Add(key,value)
	}

	encodedparams := urlparams.Encode()
	ReqURL := AccountsUri+"/oauth/v2/token?"+encodedparams

	request,reqerror := http.NewRequest("POST",ReqURL,nil)
	if reqerror != nil {
		panic(reqerror)
	}
	client := &http.Client{}
	response, resperror := client.Do(request)
    	if resperror != nil {
        		panic(resperror)
    	}

    	defer response.Body.Close()
    	respbody, _ := ioutil.ReadAll(response.Body)

	var jsonresp map[string]interface{}
	if(response.StatusCode == 200){
		json.Unmarshal(respbody, &jsonresp)
		token = jsonresp["access_token"].(string)
		return
	}else{
		json.Unmarshal(respbody, &jsonresp)
		errormsg := jsonresp["error"]
		panic(errormsg)
	}
}

/*
SendRequest - Internally used for send request.

Send request to the server and get response.
*/
func SendRequest(config map[string]string, isreturn bool) (msg interface{}, SeErr *ServerException) {
	
	params := url.Values{
		"ZOHO_ACTION": {Action},
		"ZOHO_OUTPUT_FORMAT": {OutputFormat},
		"ZOHO_ERROR_FORMAT": {ErrorFormat},
		"ZOHO_API_VERSION": {ApiVersion},
	}

	if IsOAuth == false {
		params.Add("authtoken",AuthToken)		
	}

	for key, value := range config {
		params.Add(key,value)
	}

	encodedparams := params.Encode()
    	ReqURL := RequestUri+encodedparams
    
	request, reqerror := http.NewRequest("POST", ReqURL, nil)
	if reqerror != nil {
		panic(reqerror)
	}

	if IsOAuth == true {
		request.Header.Set("Authorization" , "Zoho-oauthtoken "+AuthToken)
	}
	
	client := &http.Client{}
	response, resperror := client.Do(request)
    	if resperror != nil {
        		panic(resperror)
    	}

    	defer response.Body.Close()
    	respbody, _ := ioutil.ReadAll(response.Body)

	var jsonresult map[string]interface{}

    	if(response.StatusCode == 200){
    		if(Action == "EXPORT"){
    			format := strings.ToLower(OutputFormat)
    			filename := Exportfilename+"."+format
			OutputFormat = "JSON"
			file, err := os.Create(filename)
			if err != nil {
				panic(err)
			}
			defer file.Close()
			size, err := file.WriteString(string(respbody))
			if err != nil {
				panic(err)
			}
			fmt.Printf("Result saved to the file %s and size of the file is %d\n", filename, size)
			return 
		}else{
    			json.Unmarshal(respbody, &jsonresult)
    			if(isreturn == true){
    				msg = jsonresult["response"].(map[string]interface{})["result"]
    			}else{
    				msg = ""
    			}    		
    			return 
    		}
    	}else {
    		json.Unmarshal(respbody, &jsonresult)
    		excep := &ServerException{}
    		SeErr = excep
    		SeErr.ErrorMessage = jsonresult["response"].(map[string]interface{})["error"].(map[string]interface{})["message"].(string)
    		SeErr.ErrorCode = jsonresult["response"].(map[string]interface{})["error"].(map[string]interface{})["code"].(float64)
    		SeErr.HttpStatusCode = response.StatusCode
    		SeErr.Action = Action
    		return
    	}
}

/*
PlanInfo - It contains the plan details.

Plan - The type of the user plan.

Addons - The addon details.

BillingDate - The billing date.

RowsAllowed - The total rows allowed to the user.

RowsUsed - The number of rows used by the user.

TrialAvailed - Used to identify the trial pack.

TrialPlan - The trial plan detail.

TrialStatus - The trial plan status.

TrialEndDate - The end date of the trial plan.
*/
type PlanInfo struct{

	Plan string
	Addons string
	BillingDate string
	RowsAllowed string
	RowsUsed string
	TrialAvailed string
	TrialPlan string
	TrialStatus string
	TrialEndDate string
}

/*
ShareInfo - It contains the database shared details.

AdminMembers - Owners of the database.

GroupMembers - Group Members of the database.

SharedUsers - Shared Users of the database.

UserInfo - The PermissionInfo for the shared user.

GroupInfo - The PermissionInfo for the groups.

PublicInfo - The PermissionInfo for the public link.

PrivateInfo - The PermissionInfo for the private link.
*/
type ShareInfo struct{

	AdminMembers interface{}
	GroupMembers interface{}
	SharedUsers interface{}
	UserInfo interface{}
	GroupInfo interface{}
	PublicInfo interface{}
	PrivateInfo interface{}
}

/*
ImportResult - It contains the result of an import operation.

ImportErrors - The first 100 import errors.

ColumnDetails - The column names of the imported columns.

ImportType - The type of the import operation.

Warnings - The number of rows that were imported with warnings.

SelectedColumnCount - The number of columns that were imported.See ZOHO_SELECTED_COLUMNS parameter.

SuccessRowCount - The number of rows that were imported successfully without errors.

ImportOperation - The type of import operation.

TotalColumnCount - The total columns that were present in the imported file.

TotalRowCount - The total row count in the imported file.
*/
type ImportResult struct{

	ImportErrors string
	ColumnDetails interface{}
	ImportType string
	Warnings float64
	SelectedColumnCount float64
	SuccessRowCount float64
	ImportOperation string
	TotalColumnCount float64
	TotalRowCount float64
}

/*
ServerException - ServerException is thrown if the report server has recieved the request but did not process the request due to some error.

ErrorCode - The error code sent by the server.

ErrorMessage - The error message sent by the server.

HttpStatusCode - The http status code for the request.

Action - The action to be performed over the resource specified by the uri.
*/
type ServerException struct{

	ErrorCode float64
	ErrorMessage string
	HttpStatusCode int
	Action string
}
